/*
 * Node.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Please do not redistribute without obtaining permission.
 */

package nodenet;

//import com.sun.java.swing.*;
import javax.swing.*;
import java.util.*;
import java.io.*;
import java.awt.*;

import java.beans.PropertyVetoException;
import java.awt.image.BufferedImage;
import java.awt.font.FontRenderContext;
import java.awt.font.TextLayout;
import java.awt.geom.Rectangle2D;


/**
 * This class implements NodeNetElement.  It is responsible for the
 * GUI and maintains its own thread.  It relies on a class that
 * implements NodeBehavior to do anything.  It repeatedly calls that
 * class's act method from within a <code>while(true) {}
 * loop.</code></p>
 *
 * <p>Note: the following methods are synchronized to prevent addition
 * or subtraction of a channel during
 * {@link nodenet.NodeBehavior#transmitPacket(InputChannelVector, OutputChannelVector)}
 *
 * <ul>
 *   <li><code>public void addInputChannel(InputChannel c)</code>
 *   <li><code>public void removeInputChannel(InputChannel c)</code>
 *   <li><code>public void addOutputChannel(OutputChannel c)</code>
 *   <li><code>public void removeOutputChannel(OutputChannel c)</code>
 *   <li><code>public void run()</code>
 *   <li>
 * </ul>
 *
 * This is a change from earlier versions of nodeNet where students
 * could have written code that never returned from the transmitPacket call
 * (formerly known as act).
 *
 * @see NodeNetElement
 * @see NodeBehavior
 * @author Todd C. Parnell, tparnell@ai.mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: Node.java,v 1.15 2004/01/14 21:43:17 gus Exp $
 */
public class Node implements NodeNetElement, Resettable, Serializable {
    
    /** Diameter of the circle used to paint nodes. */
    public final static int GUISIZE = 20;
    public final static Color HIGHLIGHT_COLOR = Color.black;
    
    private final static BufferedImage dummyImage = makeImage(1,1);
    
    // This is needed for the stupid image observer stuff.
    private final static Panel dummyPanel = new Panel();
    
    private static Color[] COLORS = {
        Color.black, Color.blue, Color.gray, Color.green, Color.red,
        Color.yellow, Color.pink
    };
    
    /**
     * mapping between NodeBehaviors and Colors.  needs to be static so
     * Nodes can be serialized.
     */
    private static Hashtable nbToColorMap;
    static { Node.nbToColorMap = new Hashtable(); }
    
    final static int SLEEP_INTERVAL = 100;
    final static int DISABLED_SLEEP_INTERVAL = 1000;
    
    private static int ID;
    transient private final int myID = Node.ID++;
    
    transient private Thread spirit;
    
    
    transient private Color myColor;
    private Point pos;
    transient boolean selected;             // highlighted in GUI?
    private boolean enabled = true;         // calling act?
    private boolean destroyed;              // see note below
    private boolean stopped = true;         // signal to Thread in run()
    private String name;
    transient private NodeBehavior myBehavior;
    private String myBehaviorStr;
    private InputChannelVector inputChannels;
    private OutputChannelVector outputChannels;
    
    private NodeConfigurationBean myConfig;
    
    transient private Object lock = new Object(); // see note below
    
    //
    // Static methods
    //
    
    /**
     * Returns the current Color associated with the given NodeBehavior.
     * If a mapping does not currently exist, a new one is created.
     *
     * @param behavior the NodeBehavior whose Color will be returned
     */
    public static Color getColorFromBehavior(String behavior) {
        if (!Node.nbToColorMap.containsKey(behavior)) {
            Node.nbToColorMap.put(behavior,
            Node.COLORS[Node.nbToColorMap.size()]);
        }
        return (Color)Node.nbToColorMap.get(behavior);
    }
    
    /**
     * Gets an Enumeration of all currently known NodeBehaviors.
     */
    public static Enumeration getBehaviors() {
        return Node.nbToColorMap.keys();
    }
    
    private static BufferedImage makeImage(int x, int y) {
        return     GraphicsEnvironment.
        getLocalGraphicsEnvironment().
        getDefaultScreenDevice().
        getDefaultConfiguration().
        createCompatibleImage(x,y);
        
    }
    
    //
    // Constructors
    //
    
    public Node(NodeBehavior nb) {
        this.myConfig = new NodeConfigurationBean();
        
        myConfig.getRegistrar().nameAndRegister(this);
        
        this.name = "*UNNAMED NODE*";
        this.inputChannels = new InputChannelVector(this);
        this.outputChannels = new OutputChannelVector(this);
        this.myColor = getColorFromBehavior(nb.getClass().getName());
        this.myBehavior = nb;
        this.myBehaviorStr = this.myBehavior.getClass().getName();
    }
    
    /**
     * Build a node setting various properties. This is used when recreating
     * a network from a saved template. ID can't be set so a map of saved ID
     * to created nodes must be kept during parsing so that channels can find
     * their nodes.
     */
    public Node(Class nbc, int x, int y, String name, boolean showName,
    boolean enabled) {
        NodeBehavior nb = null;
        try {
            nb = (NodeBehavior)nbc.newInstance();
        } catch (InstantiationException ie) {
            // this should never happen, due to constructor checks
            throw new RuntimeException("InstantiationException in" +
            " createNodeType.newInstance()");
        } catch (IllegalAccessException iae) {
            // should never happen, due to constructor checks
            throw new RuntimeException("IllegalAccessException in" +
            " createNodeType.newInstance()");
        }
        
        this.myConfig = new NodeConfigurationBean();
        
        myConfig.getRegistrar().nameAndRegister(this);
        
        this.name = "*UNNAMED NODE*";
        this.inputChannels = new InputChannelVector(this);
        this.outputChannels = new OutputChannelVector(this);
        this.myColor = getColorFromBehavior(nb.getClass().getName());
        this.myBehavior = nb;
        this.myBehaviorStr = this.myBehavior.getClass().getName();
        setPos(x, y);
        setName(name);
        this.myConfig.setDisplayNodeName(showName);
    }
    
    //
    // Public methods
    //
    
    public void setPos(Point p) { this.pos = p; }
    public void setPos(int a, int b) { this.setPos(new Point(a, b)); }
    public Point getPos() { return(this.pos); }
    
    public void setSelected(boolean newState) { this.selected = newState; }
    public boolean isSelected() { return(this.selected); }
    
    public void setEnabled(boolean b) { this.enabled = b; }
    public boolean isEnabled() { return this.enabled; }
    
    public NodeConfigurationBean getConfig() { return myConfig; }
    
    public void setName(String n) {
        try {
            myConfig.setNodeName(n);
        } catch (PropertyVetoException pve) {
            JOptionPane.showMessageDialog(null,
            "Invalid Name:" + pve.getMessage(),
            "Node Naming Error",
            JOptionPane.ERROR_MESSAGE);
        }
    }
    
    public String getName() {
        return myConfig.getNodeName();
    }
    
    public boolean contains(Point p) { return(this.contains(p.x, p.y)); }
    public boolean contains(int x, int y) {
        return ((x > (this.pos.x - GUISIZE)) && (x < (this.pos.x + GUISIZE)) &&
        (y > (this.pos.y - GUISIZE)) && (y < (this.pos.y + GUISIZE)));
    }
    
    public boolean isEditable() { return false; }
    
    
    public void configure() {
        Object[] concurrentMessage =
        { "Concurrent modification detected!",
          "Please try again."};
          
          boolean configured = false;
          while (!configured) {
              myConfig.refreshComponents();
              //System.out.println(myConfig.getNodeName());
              int result = JOptionPane.showConfirmDialog(null,myConfig,"Node Configuration",JOptionPane.OK_CANCEL_OPTION);
              if (result == JOptionPane.OK_OPTION) {
                  try {
                      myConfig.implementUserChanges();
                      configured = true;
                  } catch (PropertyVetoException pve) {
                      JOptionPane.showMessageDialog(null, pve.getMessage());
                  } catch (ConcurrentModificationException cme) {
                      JOptionPane.showMessageDialog(null, concurrentMessage);
                  }
              } else {
                  configured = true;
              }
          }
    }
    
    
    /**
     * Adds a new InputChannel.  If the specified InputChannel already
     * exists (as determined by the <code>equals</code> method), it is
     * not added again.  This condition is silently ignored.
     *
     * @param c the InputChannel to connect
     */
    public void addInputChannel(InputChannel c) {
        synchronized (this.lock) {
            if (!this.inputChannels.contains(c)) {
                this.inputChannels.addElement(c);
            }
        }
    }
    
    /**
     * Removes an InputChannel.  If the specified InputChannel is not a
     * current InputChannel for this Node (as determined by the
     * <code>equals</code> method), this method does nothing.
     *
     * @param c the InputChannel to disconnect
     */
    public void removeInputChannel(InputChannel c) {
        synchronized (this.lock) {
            boolean b = this.inputChannels.removeElement(c);
        }
    }
    
    /**
     * Adds a new OutputChannel.  If the specified OutputChannel already
     * exists (as determined by the <code>equals</code> method), it is
     * not added again.  This condition is silently ignored.
     *
     * @param c the OutputChannel to connect
     */
    public void addOutputChannel(OutputChannel c) {
        synchronized (this.lock) {
            if (!this.outputChannels.contains(c)) {
                this.outputChannels.addElement(c);
            }
        }
    }
    
    /**
     * Removes an OutputChannel.  If the specified OutputChannel is not a
     * current OutputChannel for this Node (as determined by the
     * <code>equals</code> method), this method does nothing.
     *
     * @param c the OutputChannel to disconnect
     */
    public void removeOutputChannel(OutputChannel c) {
        synchronized (this.lock) {
            boolean b = this.outputChannels.removeElement(c);
        }
    }
    
     /*
      * destroy() and notifyOfDestruction() use the destroyed field to
      * prevent cycles of destruction notifications to occur.  Use of the
      * field feels like a hack, but due to the distributed nature of
      * this app, is a good solution.
      */
    
    /**
     * Destroy the Node.  <code>destroy()</code> should be called
     * whenever a Node will no longer be used.  When destroyed, Nodes
     * notify all connected input and output channels of the impending
     * destruction and remove refrences to those channels.  Typically,
     * this will force their destruction as well.  <I>Note:
     * <code>Node.finalize()</code> automatically calls
     * <code>destroy()</code></I>.
     */
    public void destroy() {
        this.stop();
        this.destroyed = true; //eliminate concurrency problems
        synchronized (this.lock) {
            //tell channels we're going away
            Enumeration enum = this.inputChannels.elements();
            while (enum.hasMoreElements()) {
                ((Channel)enum.nextElement()).notifyOfDestruction((NodeNetElement)this);
            }
            
            enum = this.outputChannels.elements();
            while (enum.hasMoreElements()) {
                ((Channel)enum.nextElement()).notifyOfDestruction((NodeNetElement)this);
            }
            this.myConfig.getRegistrar().unregister(this);
        }
    }
    
    public boolean isDestroyed() { return this.destroyed; }
    
    /** Getter for property myID.
     * @return Value of property myID.
     *
     */
    public int getMyID() {
        return myID;
    }
    
    /**
     * Generate an xml element representing a template of this node. Active
     * state such as packet counts should not be preserved. Only information
     * that is needed to recreate a network in it's pristine state
     * (i.e. before the user starts the simulation) should be included.
     *
     * @return String An xml representiation of the node.
     */
    public String templateXML() {
        return
        "<node id=\"node" + this.myID +
        "\" x=\"" + this.getPos().x +
        "\" y=\"" + this.getPos().y +
        "\" type=\"" + this.myBehavior.getClass().getName() +
        "\" enabled=\"" + this.enabled +
        "\" name=\"" + this.getName() +
        "\" showName=\"" + this.getConfig().getDisplayNodeName() +
        "\"/>";
    }
    
    /**
     * Notify this that a NodeNetElement has been destroyed.  Typically,
     * this method will be called by an input or output channel
     * connected to the node.  Once notified of destruction, all
     * references to the element are removed.
     */
    public void notifyOfDestruction(NodeNetElement bse) {
        // first see whether we've already been destroyed
        if (this.destroyed) return;
        if (bse instanceof InputChannel) {
            this.removeInputChannel( (InputChannel)bse );
        }
        if (bse instanceof OutputChannel) {
            this.removeOutputChannel( (OutputChannel)bse );
        }
    }
    
    /*
     * Render text onto a BufferedImage so that we can know it's size before
     * drawing it in the paint method.
     */
    private BufferedImage renderedText(String str, Font font, Color color, Color bgColor, int border) {
        Graphics2D G = dummyImage.createGraphics();
        FontRenderContext context = G.getFontRenderContext();
        TextLayout layout = new TextLayout(str,font,context);
        
        Rectangle2D textBounds = layout.getBounds();
        double width = textBounds.getWidth();
        double height = textBounds.getHeight();
        
        width += 3; // fudge factor... getBounds seems to be off by a little bit
        height++;
        
        int intWidth = new Double(width).intValue();
        int intHeight = new Double(height).intValue();
        
        //intHeight = Math.max(intHeight,new Float(layout.getAscent()+layout.getDescent()).intValue());
        
        intWidth += 2*border;
        intHeight += 2*border;
        
        BufferedImage rend = makeImage(intWidth,intHeight);
        
        G = rend.createGraphics();
        
        G.setPaint(bgColor);
        
        G.fill(new Rectangle2D.Double(0,0,rend.getWidth(),rend.getHeight()));
        
        G.setColor(color);
        G.setBackground(bgColor);
        
        layout.draw(G,border,intHeight-(border+layout.getDescent()/2));
        
        return rend;
    }
    
    public void paint(Graphics g) {
        if (this.destroyed || this.pos == null) return;
        // be polite & save current color
        Color tempColor = g.getColor();
        g.setColor(this.myColor);
        g.fillOval(this.pos.x - GUISIZE / 2, this.pos.y - GUISIZE / 2,
        GUISIZE, GUISIZE);
        if (this.selected) {
            g.setColor(HIGHLIGHT_COLOR);
            g.drawRect(this.pos.x - GUISIZE / 2, this.pos.y - GUISIZE / 2,
            GUISIZE, GUISIZE);
        }
        if (!this.enabled) {
            g.setColor(HIGHLIGHT_COLOR);
            g.drawLine(this.pos.x - GUISIZE / 2 - 5, this.pos.y - GUISIZE / 2 - 5,
            this.pos.x + GUISIZE / 2 + 5, this.pos.y + GUISIZE / 2 + 5);
        }
        
        if (this.myBehavior instanceof Counter) {
            String s = "" + ((Counter)this.myBehavior).getCount();
            g.setColor(Color.black);
            g.drawString(s, this.pos.x - GUISIZE, this.pos.y + GUISIZE);
        }
        
        if (this.myConfig.getDisplayNodeName()) {
            BufferedImage rendName = renderedText(this.getName(),
            g.getFont(),Color.black, Color.white, 3);
            int x = this.pos.x - rendName.getWidth()/2;
            int y = this.pos.y - ((GUISIZE / 2) + 1 + rendName.getHeight());
            
            g.drawImage(rendName,x,y,dummyPanel);
            
            
        }
        // restore color
        g.setColor(tempColor);
    }
    
    /**
     * Starts the node running.  Once running, a node will repeatedly
     * call the <code>act</code> method for it's associated
     * NodeBehavior.
     */
    public void start() {
        if (this.spirit == null) {
            this.stopped = false;
            this.spirit = new Thread(this);
            this.spirit.start();
        }
    }
    
    /**
     * Stops the node.
     */
    public void stop()  {
        if (this.spirit == null) return;
        this.stopped = true;  // signal spirit
        // now wait until spirit has exited before returning
        while (this.spirit.isAlive()) {
            try { Thread.sleep(SLEEP_INTERVAL); }
            catch (InterruptedException ie) {
                // Do nothing...
            }
        }
        this.spirit = null;     // help the gc
    }
    
    /**
     * Repeatedly calls the <code>transmitPacket</code> method for the
     * associated NodeBehavior.  <I>Note: This method should never be
     * called directly.  Instead, use the <code>start()</code> method.</I>
     */
    public void run() {
        while(!this.stopped) {
            try {
                if (this.enabled) Thread.sleep(SLEEP_INTERVAL);
                else Thread.sleep(DISABLED_SLEEP_INTERVAL);
            } catch(InterruptedException e) {}
            
            if (this.enabled && this.myBehavior != null) {
                synchronized (this.lock) {
                    this.myBehavior.transmitPacket(this.inputChannels, this.outputChannels);
                }
            }
        }
    } // run()
    
    public boolean equals(Object o) {
        if (!(o instanceof Node)) return false;
        Node n = (Node)o;
        return (n.myID == this.myID);
    }
    
    //
    // Protected instance methods
    //
    
    protected void finalize() throws IOException {
        this.destroy();
        
    }
    
    //
    // Private methods
    //
    
    /**
     * Custom de-serialization.  Get NodeBehavior & color, then start
     * Thread.
     */
    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        in.defaultReadObject();
        this.lock = new Object();
        try {
            this.myBehavior =
            (NodeBehavior)Class.forName(this.myBehaviorStr).newInstance();
            this.myColor = Node.getColorFromBehavior(this.myBehaviorStr);
        } catch (Exception e) {
            System.err.println("Error getting specified behavior. '" +
            this.myBehaviorStr + "'.");
        }
        try {
            this.myConfig.getRegistrar().register(this);
        } catch (IllegalArgumentException iae) {
            JOptionPane.showMessageDialog(null,"The name "+myConfig.getNodeName()+" exists, renaming the new one");
            this.myConfig.getRegistrar().nameAndRegister(this);
        }
    }
    
    public void reset() {
        this.selected = false;
        if (this.myBehavior instanceof Counter) {
            ((Counter)this.myBehavior).resetCount();
        }
        if (this.myBehavior instanceof Resettable) {
            ((Resettable)this.myBehavior).reset();
        }
    }
    
    
} // class Node

/*
 * $Log: Node.java,v $
 * Revision 1.15  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.14  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.13  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.12  2004/01/13 19:35:27  gus
 * Simulation Panel, channel and node are all Resettable now.
 * CountingNodeBehavior refactored into Counter
 *
 * Revision 1.11  2004/01/08 22:48:08  gus
 * Can now read xml template file (only tested breifly)
 *
 * Revision 1.10  2004/01/08 19:35:09  gus
 * further tweaks to get the xml template to contain type info.
 *
 * Revision 1.9  2004/01/08 15:46:50  gus
 * eliminate needless color argument in constructor.
 * Generate xml template
 * Convenience constructor to match template.
 *
 * Revision 1.8  2003/02/25 21:47:42  gus
 * register nodes as they are deserialized, and unregister nodes when they are destroyed
 *
 * Revision 1.7  2003/02/24 16:06:16  gus
 * let the world find out what our configuration is.
 *
 * Revision 1.6  2003/02/24 16:04:15  gus
 * Make input and output channel vectors aware of their owners, and able to report
 * who their owner is.
 *
 * Revision 1.5  2003/02/21 22:41:54  gus
 * Nodes now have configureable displayable names
 *  Modified Files:
 *  	nodenet/Node.java nodenet/NodeConfigurationBean.form
 *  	nodenet/NodeConfigurationBean.java
 *  	nodenet/registrar/UniqueNodeNamePolicy.java
 *
 * Revision 1.4  2003/02/21 18:15:36  gus
 * Reformat and make use of NodeConfigurationBean
 *
 * Revision 1.3  2003/02/14 22:19:47  gus
 * reformatted
 *
 * Revision 1.2  2002/06/13 19:32:43  gus
 * Commented out com.sun.java.swing package references, and updated obsolete
 * imorts and references into cs101 package.
 *
 * Revision 1.1  2002/06/13 17:33:21  gus
 * Moved all java files into the nodenet directory (who let them out anyway?)
 *
 * Revision 1.1.1.1  2002/06/05 21:56:35  root
 * CS101 comes to Olin finally.
 *
 * Revision 1.4  2000/05/09 06:03:54  mharder
 * Changed packagename from nodeNet to nodenet.
 *
 * Revision 1.3  1999/08/04 09:08:53  jsmthng
 * Added javadoc comments to InputChannelVector and OutputChannelVector;
 * finished updating the rest of the nodeNet package to reflect new
 * changes in name and code.
 *
 * Modified index.html to reflect the new nodeNet code, as well as to
 * clarify some parts of the problem set.
 *
 * Revision 1.1  1999/07/30 01:09:22  jsmthng
 * Renaming BinSort package to nodeNet; moving directories and files as
 * necessary.
 *
 * Revision 1.7  1999/01/21 20:57:06  tparnell
 * it worked
 *
 * Revision 1.6  1998/08/12 19:29:38  tparnell
 * Another pass after comments from las & natashao.  Added support to
 * dynamically add NodeBehaviors.  Add keyboard shortcuts to menus.  Added
 * workaround to jdk bug relating to lighweight components.  Misc other
 * bugfixes.
 *
 * Revision 1.3  1998/08/10 17:45:50  tparnell
 * Revision to use JDK1.2 and Swing.  Redesign of GUI.  Removed old kludge
 * for file I/O and replaced with object serialization.  Channel no longer
 * requires animacy.  Removed unnessary dependencies between classes.
 * Added ability to configure channel's latency and capacity.  Added
 * javadoc to all files.  General code cleanup.
 *
 */



