/*
 * NodeConfigurationBean.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Created on February 20, 2003, 5:05 PM
 */

package nodenet;

import nodenet.registrar.NodeNameRegistry;
import nodenet.registrar.NodeNameRegistrar;
import nodenet.registrar.DestinationValidator;

import java.io.Serializable;

import java.util.ConcurrentModificationException;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeSupport;
import java.beans.VetoableChangeSupport;
import java.beans.PropertyVetoException;
import java.beans.PropertyChangeListener;
import java.beans.VetoableChangeListener;

import javax.swing.JPanel;


/**
 * A bean to manage the configuration of nodes. Currently nodes can be
 * configured to have custom names, display or not display their names,
 * and be configured to hold the name of a destination (for use with
 * node behaviors that generate packets).
 *
 * @author  Patrick G. Heck gus.heck@olin.edu
 * @version $Id: NodeConfigurationBean.java,v 1.12 2004/01/14 22:10:34 gus Exp $
 */
public class NodeConfigurationBean extends JPanel implements Serializable {
    
    public static final String PROP_NODE_NAME = "Node Name";
    public static final String PROP_DISPLAY_NODE_NAME = "Display Node Name";
    public static final String PROP_DEST = "Destination";
    
    private static NodeNameRegistry registrar = NodeNameRegistry.getInstance();
    
    // Constrained:
    private String nodeName;
    private String dest;
    
    //Bound:
    private boolean displayNodeName;
    
    // Write only property: Our registrar should never be changed
    // private NodeNameRegistrar
    
    private PropertyChangeSupport propertySupport;
    private VetoableChangeSupport vetos;
    
    private DestinationValidator destVal;
    
    /** Creates new NodeConfigurationBean */
    public NodeConfigurationBean() {
        propertySupport = new PropertyChangeSupport( this );
        vetos = new VetoableChangeSupport(this);
        addVetoableChangeListener(registrar);
        destVal = DestinationValidator.getInstance();
        addVetoableChangeListener(destVal);
        initComponents();
        refreshComponents();
    }
    
    /** This method is called from within the constructor to
     * initialize the form.
     * WARNING: Do NOT modify this code. The content of this method is
     * always regenerated by the Form Editor.
     */
    private void initComponents() {//GEN-BEGIN:initComponents
        java.awt.GridBagConstraints gridBagConstraints;

        namePanel = new javax.swing.JPanel();
        nameLabel = new javax.swing.JLabel();
        nameTextField = new javax.swing.JTextField();
        nameDisplayCheckBox = new javax.swing.JCheckBox();
        destinationPanel = new javax.swing.JPanel();
        destinationLabel = new javax.swing.JLabel();
        destinationTextField = new javax.swing.JTextField();
        destDetailLabel = new javax.swing.JLabel();

        setLayout(new javax.swing.BoxLayout(this, javax.swing.BoxLayout.Y_AXIS));

        namePanel.setLayout(new java.awt.GridBagLayout());

        nameLabel.setText("Node name:");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.weightx = 1.0;
        gridBagConstraints.insets = new java.awt.Insets(5, 5, 5, 5);
        namePanel.add(nameLabel, gridBagConstraints);

        nameTextField.setColumns(15);
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 0;
        namePanel.add(nameTextField, gridBagConstraints);

        nameDisplayCheckBox.setText("Display Node Name");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.insets = new java.awt.Insets(0, 5, 5, 5);
        namePanel.add(nameDisplayCheckBox, gridBagConstraints);

        add(namePanel);

        destinationPanel.setLayout(new java.awt.GridBagLayout());

        destinationLabel.setText("Packet Destination:");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.anchor = java.awt.GridBagConstraints.WEST;
        gridBagConstraints.weightx = 1.0;
        destinationPanel.add(destinationLabel, gridBagConstraints);

        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 1;
        gridBagConstraints.gridy = 0;
        gridBagConstraints.fill = java.awt.GridBagConstraints.HORIZONTAL;
        gridBagConstraints.weightx = 1.0;
        destinationPanel.add(destinationTextField, gridBagConstraints);

        destDetailLabel.setFont(new java.awt.Font("Dialog", 0, 10));
        destDetailLabel.setText("(Not yet implemented)");
        gridBagConstraints = new java.awt.GridBagConstraints();
        gridBagConstraints.gridx = 0;
        gridBagConstraints.gridy = 1;
        gridBagConstraints.gridwidth = 2;
        gridBagConstraints.weightx = 1.0;
        destinationPanel.add(destDetailLabel, gridBagConstraints);

        add(destinationPanel);

    }//GEN-END:initComponents
    
    /**
     * Gets the identifying name of the node.
     *
     * @return The name of the node.
     */
    public String getDest() {
        return dest;
    }
    
    /**
     * Attempts to change the neme of the node. If the name is not a
     * legal name it throws an exception and the name will remain unchanged.
     *
     * @param value The desired new name for the node
     * @throws PropertyVetoException if the NodeNameRegistry is implementing
     *         a registration policy that forbids the name.
     * @see NodeNameRegistry
     * @see NodeNameRegistrar
     */
    public void setDest(String value) throws PropertyVetoException {
        String oldValue = dest;
        synchronized (registrar) {
            vetos.fireVetoableChange(PROP_DEST, oldValue, value);
            dest = value;
        }
        propertySupport.firePropertyChange(PROP_DEST, oldValue, dest);
    }
    
    
    /**
     * Gets the identifying name of the node.
     *
     * @return The name of the node.
     */
    public String getNodeName() {
        return nodeName;
    }
    
    /**
     * Attempts to change the neme of the node. If the name is not a
     * legal name it throws an exception and the name will remain unchanged.
     *
     * @param value The desired new name for the node
     * @throws PropertyVetoException if the NodeNameRegistry is implementing
     *         a registration policy that forbids the name.
     * @see NodeNameRegistry
     * @see NodeNameRegistrar
     */
    public void setNodeName(String value) throws PropertyVetoException {
        String oldValue = nodeName;
        synchronized (registrar) {
            vetos.fireVetoableChange(PROP_NODE_NAME, oldValue, value);
            nodeName = value;
        }
        propertySupport.firePropertyChange(PROP_NODE_NAME, oldValue, nodeName);
    }
    
    /**
     * A convenience method that yeilds the same result as
     * NodeNameRegistry.getInstance().
     *
     * @return The NodeNameRegistry of which there can be only one.
     * @see NodeNameRegistry
     */
    public NodeNameRegistrar getRegistrar() {
        return registrar;
    }
    
    /**
     * Display the name of the node or not.
     *
     * @return <code>true</code> if the name should be displayed <code>false</code>
     *         otherwise.
     */
    
    public boolean getDisplayNodeName() {
        return displayNodeName;
    }
    
    /**
     * Sets whether or not the name of the node is displayed.
     *
     * @param display whether or not to display the node.
     */
    public void setDisplayNodeName(boolean display) {
        boolean oldVal = getDisplayNodeName();
        displayNodeName = display;
        propertySupport.firePropertyChange(PROP_DISPLAY_NODE_NAME, oldVal, display);
    }
    
    /**
     * Register an object to be notified when the properties of this bean
     * are changed.
     *
     * @param listener The listening object
     */
    public void addPropertyChangeListener(PropertyChangeListener listener) {
        propertySupport.addPropertyChangeListener(listener);
    }
    
    /**
     * Unregister an object that nolonger needs to be notified of property
     * changes.
     *
     * @param listener The object that wants to stop listening.
     */
    public void removePropertyChangeListener(PropertyChangeListener listener) {
        propertySupport.removePropertyChangeListener(listener);
    }
    
    
    /**
     * Register an object that has the power to overide changes to this bean.
     *
     * @param l An object that can veto changes.
     */
    public void addVetoableChangeListener(VetoableChangeListener l) {
        vetos.addVetoableChangeListener(l);
    }
    
    /**
     * Unregister an object that nolonger needs to overide changes to this bean.
     *
     * @param l A vetoer that nolonger wishes to vote on changes.
     */
    public void removeVetoableChangeListener(VetoableChangeListener l) {
        vetos.removeVetoableChangeListener(l);
    }
    
    /**
     * Refresh the values displayed in the components used to display and
     * configure this bean. This should be called immediately before displaying
     * the bean in a Container.
     */
    public synchronized void refreshComponents() {
        
        // until destinations are implemented
        this.destinationPanel.setVisible(false);
        
        this.nameTextField.setText(this.getNodeName());
        this.nameDisplayCheckBox.setSelected(this.getDisplayNodeName());
        this.destinationTextField.setText(this.getDest());
    }
    
    /**
     * Attempt to enact user changes. This method is typically called after
     * the panel has been shown in a dialog and properties may have been
     * edited. Changes are applied atomically once the have all been validated.
     * If any changes have been vetoed, this method will throw the related
     * exception. If conditions change after vetos have been checked causing
     * a subsequent property veto a ConcurrentModificationException will be
     * thrown.
     *
     * @throws PropertyVetoException if listenerns object.
     */
    public synchronized void implementUserChanges() throws PropertyVetoException,
    ConcurrentModificationException {
        
        boolean reset=false;
        
        // gather a snapshot of the data from the components and the bean
        String newName = nameTextField.getText();
        String oldName = this.getNodeName();
        boolean newShowName = nameDisplayCheckBox.isSelected();
        boolean oldShowName = this.getDisplayNodeName();
        String newDest = destinationTextField.getText();
        String oldDest = this.getDest();
        
        // test the changes so that we throw a PropertyVetoException in
        // preference to the ConcurrentModificationException. PVE's have more
        // information that can be used to clarify the problem to the user.
        // This also avoids having to re-throw the exception after we reset
        // any values that have been changed.
        if (!newName.equals(this.getNodeName())) {
            vetos.fireVetoableChange(PROP_NODE_NAME, oldName, newName);
        }
        vetos.fireVetoableChange(PROP_DEST, oldDest, newDest);
        // Now try to make the changes that can be vetoed.
        try {
            setNodeName(newName);
            setDest(newDest);
        } catch (PropertyVetoException pve) {
            reset = true;
        }
        
        // if something failed we have been concurrently modified,
        // revert and throw and exception.
        if (reset) {
            nodeName=oldName;
            dest = oldDest;
            throw new ConcurrentModificationException("Node configuration" +
            " concurrently modified");
        }
        
        // Now proceed with non-vetoable changes
        setDisplayNodeName(newShowName);
    }
    
    
    
    
    // Variables declaration - do not modify//GEN-BEGIN:variables
    private javax.swing.JLabel destDetailLabel;
    private javax.swing.JLabel destinationLabel;
    private javax.swing.JPanel destinationPanel;
    private javax.swing.JTextField destinationTextField;
    private javax.swing.JCheckBox nameDisplayCheckBox;
    private javax.swing.JLabel nameLabel;
    private javax.swing.JPanel namePanel;
    private javax.swing.JTextField nameTextField;
    // End of variables declaration//GEN-END:variables
    
}

/*
 * $Log: NodeConfigurationBean.java,v $
 * Revision 1.12  2004/01/14 22:10:34  gus
 * hide the destination panel until such time as we are ready to use it.
 *
 * Revision 1.11  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.10  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.9  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.8  2004/01/06 18:50:37  gus
 * note that packet routing is not implemented
 *
 * Revision 1.7  2004/01/06 17:09:19  gus
 * Netbeans generated code changed
 *
 * Revision 1.6  2003/02/25 22:35:18  gus
 * adding log comment
 *
 */
