/*
 * NodeNetElement.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Please do not redistribute without obtaining permission.
 */

package nodenet;

import java.awt.Graphics;
import java.awt.Point;

/**
 * NodeNetElement is an interface implemented by any and all Nodes
 * and channels.  It defines the basic methods any element must
 * have.<p>
 *
 * @author Todd C. Parnell, tparnell@ai.mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: NodeNetElement.java,v 1.5 2004/01/14 21:43:17 gus Exp $
 */
public interface NodeNetElement extends Runnable {
    
    /**
     * Paint ourself to a graphics context.
     *
     * @param g The graphics context on which to render ourself
     */
    public void paint(Graphics g);
    
    /**
     * Notify this object that it should render itself as selected or
     * unselected.
     *
     * @param b the new selection state
     */
    public void setSelected(boolean b);
    
    /**
     * Determine the selection stae of this element.
     *
     * @return the current selection state
     */
    public boolean isSelected();
    
    /**
     * Set the enabled property. This property determines whether or not the
     * element participates in the simulation. Elements that are disabled
     * should never have any behavior beyond the fact that they exist. Enabled
     * elements interact with other elements by generating, absorbing or
     * transmitting packets.
     *
     * @param b the new state of the enabled property
     */
    public void setEnabled(boolean b);
    
    /**
     * Determine if this element is enabled.
     *
     * @see #setEnabled(boolean)
     * @return the current state of the enabled property.
     */
    public boolean isEnabled();
    
    /**
     * Determine if this element contains the specified point.
     *
     * @param p the point to test
     * @return true if the point is contained, false otherwise.
     */
    public boolean contains(Point p);
    
    
    /**
     * Determine if this element contains the specified point.
     *
     * @param x the x coordinate of the point to test
     * @param y the y coordinate of the point to test
     * @return true if the point is contained, false otherwise.
     */
    public boolean contains(int x, int y);
    
    /**
     * Display a configuratin dialog to the user and record changes made
     * to the configuration via this dialog.
     */
    public void configure();
    
    /**
     * Set the name of this instance. Names may or may not be unique, and may or
     * may not play a role in the functionality of the class.
     *
     * @param name The name to use for this element.
     */
    public void setName(String name);
    
    /**
     * Get the name of this instance.
     *
     * @see #setName(String)
     * @return then name of this element
     */
    public String getName();
    
    /**
     * begin exhibiting the behavior of this element This method
     * is called when a simulation is started.
     */
    public void start();
    
    /**
     * Stop exhibiting the behavior of this element. This method is called
     * on all NodeNetElements withing a simulation pannel when a
     * simulation is stopped.
     */
    public void stop();
    
    /**
     * Permanantly release all resources held by this element, and allow it's
     * thread to run to completion. This must be called to make an element
     * eligable for garbage collection
     */
    public void destroy();
    
    /**
     * Test to see if this element has been destroyed.
     *
     * @return true if the object is destroyed false otherwise.
     */
    public boolean isDestroyed();
    
    /**
     * Indicate that another element referenced by this object is being
     * destroyed, and references to the specified element should be dropped.
     *
     * @param nne the element that is being destroyed
     */
    public void notifyOfDestruction(NodeNetElement nne);
}

/*
 * $Log: NodeNetElement.java,v $
 * Revision 1.5  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.4  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.3  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.2  2004/01/13 19:35:27  gus
 * Simulation Panel, channel and node are all Resettable now.
 * CountingNodeBehavior refactored into Counter
 *
 * Revision 1.1  2002/06/13 17:33:21  gus
 * Moved all java files into the nodenet directory (who let them out anyway?)
 *
 * Revision 1.1.1.1  2002/06/05 21:56:35  root
 * CS101 comes to Olin finally.
 *
 * Revision 1.4  2000/05/09 06:03:54  mharder
 * Changed packagename from nodeNet to nodenet.
 *
 * Revision 1.3  1999/08/04 09:08:53  jsmthng
 * Added javadoc comments to InputChannelVector and OutputChannelVector;
 * finished updating the rest of the nodeNet package to reflect new
 * changes in name and code.
 *
 * Modified index.html to reflect the new nodeNet code, as well as to
 * clarify some parts of the problem set.
 *
 * Revision 1.1  1999/07/30 01:09:23  jsmthng
 * Renaming BinSort package to nodeNet; moving directories and files as
 * necessary.
 *
 * Revision 1.3  1998/08/12 19:29:25  tparnell
 * Another pass after comments from las & natashao.  Added support to
 * dynamically add NodeBehaviors.  Add keyboard shortcuts to menus.  Added
 * workaround to jdk bug relating to lighweight components.  Misc other
 * bugfixes.
 *
 * Revision 1.3  1998/08/10 17:45:50  tparnell
 * Revision to use JDK1.2 and Swing.  Redesign of GUI.  Removed old kludge
 * for file I/O and replaced with object serialization.  Channel no longer
 * requires animacy.  Removed unnessary dependencies between classes.
 * Added ability to configure channel's latency and capacity.  Added
 * javadoc to all files.  General code cleanup.
 *
 */
