/*
 * SimulationPanel.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Please do not redistribute without obtaining permission.
 */

package nodenet;

import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.EntityResolver;

import java.io.IOException;
import java.io.InputStream;
import java.io.Serializable;
import java.io.StringReader;
import java.io.ObjectInputStream;
import java.io.InputStreamReader;

import java.awt.Point;
import java.awt.Graphics;
import java.awt.Rectangle;
import java.awt.Dimension;

import java.awt.event.MouseEvent;
import java.awt.event.MouseAdapter;

import java.beans.PropertyChangeListener;

import java.util.List;
import java.util.Vector;
import java.util.Iterator;
import java.util.Enumeration;
import java.util.StringTokenizer;

import javax.xml.parsers.DocumentBuilder;
import javax.xml.parsers.DocumentBuilderFactory;
import javax.xml.parsers.ParserConfigurationException;

import cs101.xml.SimpleErrorHandler;

import javax.swing.JPanel;

/**
 * A graphical display of a group of NodeNetElements.  It manages
 * construction and destruction, as well as placement and
 * configuration of individual elements.
 *
 * <p>A SimulationPanel listens to the selectedBehavior Property on
 * a NodeTypeSelector (or any other future class that provides behavior
 * selection functionality). The cached value from the last property change
 * event is used to place new nodes when the mouse is clicked.
 *
 * @see NodeNetElement
 *
 * @author Todd C. Parnell, tparnell@ai.mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: SimulationPanel.java,v 1.17 2004/01/16 20:48:59 gus Exp $
 */
class SimulationPanel extends JPanel implements Runnable, Serializable,
Resettable, PropertyChangeListener {
    
    /** dimensions */
    private static int WIDTH  = 400;
    private static int HEIGHT = 350;
    
    /**
     * For serialization.  Change to a different number whenever
     * incompatible changes are made.
     */
    static final long serialVersionID = 1L;
    
    /** used for equals() */
    private static int ID_GENERATOR = 0;
    /** a unique identifier among all SimulationPanels in the VM */
    transient private final int ID = ID_GENERATOR++;
    /** The elements we control */
    private List bsElements;
    /** flag to indicate whether the simulation (not the thread) is running */
    transient boolean running;
    /** element that is graphically selected */
    transient private NodeNetElement selectedNNE = null;
    /** the animacy */
    transient private Thread spirit;
    /** flag to the Thread */
    private boolean stopped;
    /** how often to repaint */
    private final int repaint_pause = 100;
    /** when creating a Channel, the starting Node */
    transient private Node anchor;
    /** defines the type of node to create when the mouse is clicked. */
    private Class createNodeType;
    /** remember who told us what node type to use */
    private NodeBehaviorProvider behaviorSource;
    
    
    
    
    /**
     * Creates a new, empty SimulationPanel.  SimulationPanels start with
     * the simulation off.
     *
     */
    public SimulationPanel() {
        super();
        
        this.bsElements = new Vector();
        this.initListeners();
        this.spirit = new Thread(this);
        this.spirit.start();
    }
    
    private void initListeners() {
        this.addMouseListener(new MouseAdapter() {
            public void mouseClicked(MouseEvent e) {
                //System.out.println("Mouse Event:" + e );
                Point temp = e.getPoint();
                NodeNetElement nne = null;
                try {
                    nne = whatsInside(temp);
                    SimulationPanel.this.setSelected(nne);
                    // if we get here, bse is non-null
                    if (SimulationPanel.this.anchor != null) {
                        if (nne instanceof Channel) {
                            // oops.  can't connect these up.
                            SimulationPanel.this.anchor = null;
                            return;
                        }
                        // connect the dots with a new channel
                        Node n = (Node)nne;
                        if (channelExists(SimulationPanel.this.anchor, n)) return;
                        try {
                            Channel bsc = new Channel(SimulationPanel.this.anchor, n);
                            SimulationPanel.this.bsElements.add(bsc);
                            SimulationPanel.this.setSelected(bsc);
                            if (SimulationPanel.this.running) bsc.start();
                            return;
                        } catch (SameNodeException sne) {
                        } finally {
                            SimulationPanel.this.anchor = null;
                        }
                    }
                    // if the user has double clicked on a node, set the anchor
                    if (e.getClickCount() > 1 && nne instanceof Node) {
                        SimulationPanel.this.anchor = (Node)nne;
                        return;
                    }
                }
                catch (IsInsideNothingException iine) {
                    //System.out.println("not inside");
                    // user clicked on empty space
                    if (SimulationPanel.this.anchor != null) {
                        // remove old anchor
                        SimulationPanel.this.anchor = null;
                        return;
                    } else if (SimulationPanel.this.createNodeType != null) {
                        // create a new node in this location
                        SimulationPanel.this.anchor = null;
                        NodeBehavior nb = null;
                        try {
                            nb = (NodeBehavior)SimulationPanel.
                            this.createNodeType.newInstance();
                        } catch (InstantiationException ie) {
                            // this should never happen, due to constructor checks
                            throw new RuntimeException("InstantiationException in" +
                            " createNodeType.newInstance()");
                        } catch (IllegalAccessException iae) {
                            // should never happen, due to constructor checks
                            throw new RuntimeException("IllegalAccessException in" +
                            " createNodeType.newInstance()");
                        }
                        Node n = new Node(nb);
                        if (n == null) return;
                        SimulationPanel.this.setSelected((NodeNetElement)n);
                        n.setPos(temp);
                        if (SimulationPanel.this.running) n.start();
                        SimulationPanel.this.bsElements.add(n);
                        return;
                    }
                }
                finally {
                    SimulationPanel.this.repaint();
                }
            }
        });
    } // initListeners()
    
    public Dimension getMinimumSize() { return(this.getPreferredSize()); }
    public Dimension getPreferredSize() {
        return new Dimension(SimulationPanel.WIDTH, SimulationPanel.HEIGHT);
    }
    
    public void paint(Graphics g) {
        
        Rectangle r = g.getClipBounds();
        g.clearRect(r.x, r.y, r.width, r.height);
        
        // paint channels beneth nodes
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            NodeNetElement nne = (NodeNetElement)iter.next();
            if (nne instanceof Channel) nne.paint(g);
        }
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            NodeNetElement nne = (NodeNetElement)iter.next();
            if (nne instanceof Node) nne.paint(g);
        }
    }
    
    /**
     * Begins stepping through the simulation.  Starts all current
     * nodes and channels.
     */
    public void startSimulation() {
        if (this.running) return;
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            ((NodeNetElement)iter.next()).start();
        }
        this.running = true;
    }
    
    /**
     * Stops the current simulation.<p>
     *
     * Note: this method waits for each Node and Channel to stop their
     * own execution before completing.  If the act method of any Node
     * does not return, this method will not return either.
     */
    public void stopSimulation() {
        if (!this.running) return;
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            ((NodeNetElement)iter.next()).stop();
        }
        this.running = false;
    }
    
    /**
     * Stops the current simulation and removes all NodeNetElements.<p>
     *
     * Note: this method waits for each Node and Channel to stop their
     * own execution before completing.  If the act method of any Node
     * does not return, this method will not return either.
     */
    public void clear() {
        this.stopSimulation();
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            ((NodeNetElement)iter.next()).destroy();
        }
        this.bsElements = new Vector();
    }
    
    /**
     * Enables or disables the currently selected NodeNetElement.
     * Disabled elements will still be visible in the simulation and
     * part of it.  However, disabled elements will not move packets.
     *
     * @param state the new state of the selection
     */
    public void setSelectionEnabled(boolean state) {
        if (this.selectedNNE == null) return;
        this.selectedNNE.setEnabled(state);
    }
    
    /**
     * Deletes the currently selected NodeNetElement.  If the selected
     * element is a Node, all channels connected to it are destroyed as
     * well.
     */
    public void deleteSelection() {
        if (this.selectedNNE == null) return;
        boolean wasRunning = this.running;
        if (wasRunning) this.stopSimulation();
        this.selectedNNE.destroy();
        Vector temp = new Vector();
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            NodeNetElement nne = (NodeNetElement)iter.next();
            if (!nne.isDestroyed()) temp.add(nne);
        }
        this.bsElements = temp;
        if (wasRunning) this.startSimulation();
    }
    
    /**
     * Pops up a configuration dialog for the currently selected
     * element.
     */
    public void configureSelection() {
        if (this.selectedNNE != null) this.selectedNNE.configure();
    }
    
    /**
     * Interactive control loop.  Just calls repaint()
     */
    public void run() {
        try {
            while(!this.stopped) {
                Thread.sleep(this.repaint_pause);
                this.repaint();
            }
        }
        catch(InterruptedException e){}
    }
    
    public boolean equals(Object o) {
        if (!(o instanceof SimulationPanel)) return false;
        return (this.ID == ((SimulationPanel)o).ID);
    }
    
    public String templateXML() {
        StringBuffer buff = new StringBuffer();
        buff.append("<?xml version=\"1.0\" standalone=\"no\"?>\n");
        buff.append("<!DOCTYPE nodenet SYSTEM \"nodenet.dtd\">\n");
        buff.append("<nodenet>\n");
        for (Iterator iter = bsElements.iterator(); iter.hasNext();) {
            Object o = iter.next();
            if (o instanceof Node) {
                Node n = (Node) o;
                buff.append("  " + n.templateXML() + "\n");
            }
        }
        // all channels must come after all nodes.
        for (Iterator iter = bsElements.iterator(); iter.hasNext();) {
            Object o = iter.next();
            
            if (o instanceof Channel) {
                Channel c = (Channel) o;
                buff.append("  " + c.templateXML() + "\n");
            }
            
        }
        buff.append("</nodenet>\n");
        
        String result = buff.toString();
        
        Class[] nbehaviors = behaviorSource.getNodeBehaviors();
        result = result.replaceAll(GeneratorNodeBehavior.class.getName(), "Generator");
        result = result.replaceAll(TerminatorNodeBehavior.class.getName(), "Terminator");
        for (int i = 0; i < nbehaviors.length; i++) {
            result = result.replaceAll(nbehaviors[i].getName(), "Behavior" + i);
        }
        
        // check our work before others get to see it...
        try {
            DocumentBuilderFactory dbf = DocumentBuilderFactory.newInstance();
            dbf.setExpandEntityReferences(true);
            dbf.setValidating(true);
            
            DocumentBuilder docBuild = dbf.newDocumentBuilder();
            docBuild.setErrorHandler(new SimpleErrorHandler());
            docBuild.setEntityResolver( new DTDResolver());
            
            InputSource is = new InputSource(new StringReader(result));
            
            // it appears that crimson doesn't quite follow the javadoc spec.
            // It fails to call entity resolver if systemID is not set.
            // the javadoc indicates that entity resolver should still get
            // called, and infact override the value.... but crimson has a fit
            // and starts throwing things (MalformedURIException,
            // SaxParseException), apparently trying to build the URI
            // for SystemID and prepend it to sys.
            //
            // 5 hours of hair pulling, but finally found it on log4j's mailing
            // list (only about the 20th search I ran) So yeah, it's neccesary.
            is.setSystemId(DTDResolver.KLUDGE);
            
            docBuild.parse(is);
        } catch (ParserConfigurationException pce) {
            pce.printStackTrace();
        } catch (SAXException saxe) {
            saxe.printStackTrace();
        } catch (IOException ioe) {
            System.out.println(ioe.getMessage());
            ioe.printStackTrace();
        }
        return result;
    }
    //
    // Protected instance methods
    //
    
    protected void finalize() throws IOException {
        this.stopped = true;
        this.stopSimulation();
    }
    
    //
    // Private instance methods
    //
    
    /**
     * Switches the selected element.
     *
     * @param bse the element to select
     */
    private void setSelected(NodeNetElement nne) {
        if (this.selectedNNE != null) this.selectedNNE.setSelected(false);
        this.selectedNNE = nne;
        if (this.selectedNNE != null) this.selectedNNE.setSelected(true);
    }
    
    /**
     * Determines the element at a particular Point on the GUI.
     *
     * @param p the point on the gui of interest
     */
    private NodeNetElement whatsInside(Point p)
    throws IsInsideNothingException {
        NodeNetElement nne;
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            nne = (NodeNetElement)iter.next();
            if (nne.contains(p)) return nne;
        }
        throw new IsInsideNothingException();
    }
    
    /**
     * Determines whether a channel exists between the specified nodes.
     */
    private boolean channelExists(Node start, Node finish) {
        Channel ch;
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            NodeNetElement nne = (NodeNetElement)iter.next();
            if (!(nne instanceof Channel)) continue;
            ch = (Channel)nne;
            if (((ch.getStartNode()).equals(start)) &&
            ((ch.getEndNode()).equals(finish))) {
                return true;
            }
        }
        return false;
    }
    
    private void readObject(ObjectInputStream in)
    throws IOException, ClassNotFoundException {
        // set up fields
        in.defaultReadObject();
        // listen to mouse events
        this.initListeners();
        // start the thread
        this.spirit = new Thread(this);
        this.spirit.start();
    }
    
    public void propertyChange(java.beans.PropertyChangeEvent propertyChangeEvent) {
        //System.out.println("recieved event, New value="+propertyChangeEvent.getNewValue());
        if ("selectedBehavior".equals(propertyChangeEvent.getPropertyName())) {
            Object source = propertyChangeEvent.getSource();
            if (!(source instanceof NodeBehaviorProvider)) {
                throw new Error("Internal Program ERROR! Simulation panel is"+
                "getting selectedBehavior events from an illegal source!");
            } else {
                this.createNodeType = (Class) propertyChangeEvent.getNewValue();
                this.behaviorSource = (NodeBehaviorProvider) source;
            }
        }
    }
    
    /** Getter for property bsElements.
     * @return Value of property bsElements.
     *
     */
    public List getBsElements() {
        return bsElements;
    }
    
    /** Setter for property bsElements.
     * @param bsElements New value of property bsElements.
     *
     */
    public void setBsElements(List bsElements) {
        this.bsElements = bsElements;
    }
    
    /** Getter for property behaviorSource.
     * @return Value of property behaviorSource.
     *
     */
    public nodenet.NodeBehaviorProvider getBehaviorSource() {
        return behaviorSource;
    }
    
    /** Setter for property behaviorSource.
     * @param behaviorSource New value of property behaviorSource.
     *
     */
    public void setBehaviorSource(nodenet.NodeBehaviorProvider behaviorSource) {
        this.behaviorSource = behaviorSource;
    }
    
    /**
     * Reset all our Resettable nodenet elements.
     */
    public void reset() {
        for (Iterator iter = this.bsElements.iterator(); iter.hasNext();) {
            try {
                ((Resettable) iter.next()).reset();
            } catch (ClassCastException cce) {
                //ignore
            }
        }
    }
    
}

/*
 * $Log: SimulationPanel.java,v $
 * Revision 1.17  2004/01/16 20:48:59  gus
 * Allow user to add jars and dirs to the classpath by using a custom
 * ClassLoader.
 * Also (try to) prevent the window from getting too small since funny things
 * happen with the JSeparator and very small window sizes.
 *
 * Revision 1.16  2004/01/15 16:45:00  gus
 * still working on encoding issues
 *
 * Revision 1.15  2004/01/14 23:42:35  gus
 * specifying the encoding is clearly causing more problems that it is worth
 *
 * Revision 1.14  2004/01/14 23:28:11  gus
 * another attempt to get the encoding right
 *
 * Revision 1.13  2004/01/14 23:19:17  gus
 * Make sure the encoding matches the file, rather than assuming utf-8
 *
 * Revision 1.12  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.11  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.10  2004/01/13 19:35:27  gus
 * Simulation Panel, channel and node are all Resettable now.
 * CountingNodeBehavior refactored into Counter
 *
 * Revision 1.9  2004/01/08 22:48:08  gus
 * Can now read xml template file (only tested breifly)
 *
 * Revision 1.8  2004/01/08 19:35:09  gus
 * further tweaks to get the xml template to contain type info.
 *
 * Revision 1.7  2004/01/08 15:52:25  gus
 * DTD for nodenet.
 * generate xml and validate it against the DTD.
 *
 * Revision 1.6  2004/01/07 20:10:12  gus
 * Update comments to refelect disappearance of ControlPanel Class
 *
 * Revision 1.5  2004/01/07 19:56:17  gus
 * Another comment refereing to ControlPanel
 *
 * Revision 1.4  2004/01/07 19:55:02  gus
 * missed a reference to ControlPanel (though it was commented)
 *
 * Revision 1.3  2004/01/07 18:32:47  gus
 * Ripped out ControlPanel and NodeNetRadioButton.
 * Replaced them with NodeTypeSelector, which provides
 * The loaded User NodeBehavior types as an indexed
 * property. Simulation panel now implements PropertyChangeListener
 * to listen for changes in the selected node behavior. The
 * buttons in the display are now JToggleButtons, and will
 * use the public field BEHAVIOR_NAME to lable the buttons
 * if it is declared for any NodeBehavior.
 *
 * Revision 1.2  2002/06/13 19:32:43  gus
 * Commented out com.sun.java.swing package references, and updated obsolete
 * imorts and references into cs101 package.
 *
 * Revision 1.1  2002/06/13 17:33:21  gus
 * Moved all java files into the nodenet directory (who let them out anyway?)
 *
 * Revision 1.1.1.1  2002/06/05 21:56:35  root
 * CS101 comes to Olin finally.
 *
 * Revision 1.4  2000/05/09 06:03:54  mharder
 * Changed packagename from nodeNet to nodenet.
 *
 * Revision 1.3  1999/08/04 09:08:55  jsmthng
 * Added javadoc comments to InputChannelVector and OutputChannelVector;
 * finished updating the rest of the nodeNet package to reflect new
 * changes in name and code.
 *
 * Modified index.html to reflect the new nodeNet code, as well as to
 * clarify some parts of the problem set.
 *
 * Revision 1.1  1999/07/30 01:09:24  jsmthng
 * Renaming BinSort package to nodeNet; moving directories and files as
 * necessary.
 *
 * Revision 1.4  1999/01/21 20:57:07  tparnell
 * it worked
 *
 * Revision 1.3  1998/08/12 19:29:42  tparnell
 * Another pass after comments from las & natashao.  Added support to
 * dynamically add NodeBehaviors.  Add keyboard shortcuts to menus.  Added
 * workaround to jdk bug relating to lighweight components.  Misc other
 * bugfixes.
 *
 * Revision 1.3  1998/08/10 17:45:50  tparnell
 * Revision to use JDK1.2 and Swing.  Redesign of GUI.  Removed old kludge
 * for file I/O and replaced with object serialization.  Channel no longer
 * requires animacy.  Removed unnessary dependencies between classes.
 * Added ability to configure channel's latency and capacity.  Added
 * javadoc to all files.  General code cleanup.
 *
 */
