/*
 * TemplateParser.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Please do not redistribute without obtaining permission.
 * Created on January 8, 2004, 2:53 PM
 */

package nodenet;

import org.xml.sax.XMLReader;
import org.xml.sax.Attributes;
import org.xml.sax.InputSource;
import org.xml.sax.SAXException;
import org.xml.sax.helpers.DefaultHandler;

import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.IOException;

import java.util.List;
import java.util.Vector;
import java.util.HashMap;
import java.util.ArrayList;

import javax.xml.parsers.SAXParser;
import javax.xml.parsers.SAXParserFactory;
import javax.xml.parsers.ParserConfigurationException;

import cs101.xml.SimpleErrorHandler;




/**
 * A class for reading a template file and converting it into a List
 * of {@link NodeNetElement}.
 *
 * @author  Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: TemplateParser.java,v 1.4 2004/01/14 21:43:17 gus Exp $
 */
public class TemplateParser {
    
    private NodeBehaviorProvider provider;
    private List elements;
    private List warnings;
    
    /** Creates a new instance of TemplateParser */
    public TemplateParser(NodeBehaviorProvider provider) {
        this.provider = provider;
    }
    
    public List parse(String fName) throws FileNotFoundException,
    ParserConfigurationException, SAXException {
        this.warnings = new ArrayList();
        
        elements = new Vector();
        FileReader fr = new FileReader(fName);
        InputSource is = new InputSource(fr);
        is.setSystemId(DTDResolver.KLUDGE);
        
        SAXParserFactory spf = SAXParserFactory.newInstance();
        spf.setValidating(true);
        SAXParser parser = spf.newSAXParser();
        XMLReader xRead = parser.getXMLReader();
        xRead.setErrorHandler(new SimpleErrorHandler());
        xRead.setContentHandler(new myContentHandler());
        xRead.setEntityResolver(new DTDResolver());
        try {
            xRead.parse(is);
            fr.close();
        } catch (IOException ioe) {
            throw new SAXException(ioe);
        }
        
        return elements;
    }
    
    /**
     * Getter for property warnings.
     *
     * @return Value of property warnings.
     */
    public java.util.List getWarnings() {
        return warnings;
    }
    
    /**
     * Add a message to the warnings.
     */
    
    private void warn(String msg) {
        warnings.add(msg);
    }
    
    private class myContentHandler extends DefaultHandler {
        
        /**
         * keep track of what ID was assigned to the node created
         * and relate it to the ID in the file with this map.
         */
        HashMap nodeMap = new HashMap();
        
        public void startElement(String uri, String localName,
        String qname, Attributes attributes) throws SAXException {
            
            if ("node".equals(qname)) {
                String idVal = attributes.getValue("id");
                
                String nbVal = attributes.getValue("type");
                String xVal = attributes.getValue("x");
                String yVal = attributes.getValue("y");
                String showVal = attributes.getValue("showName");
                String enabledVal = attributes.getValue("enabled");
                
                String name = attributes.getValue("name"); // no conversion
                
                idVal = idVal.replaceFirst("node", "");
                Integer id = new Integer(idVal);
                int x = new Integer(xVal).intValue();
                int y = new Integer(yVal).intValue();
                boolean show = new Boolean(showVal).booleanValue();
                boolean enabled = new Boolean(enabledVal).booleanValue();
                
                Class nb = null;
                if ("Generator".equals(nbVal)) {
                    nb = GeneratorNodeBehavior.class;
                }
                
                if ("Terminator".equals(nbVal)) {
                    nb = TerminatorNodeBehavior.class;
                }
                
                if (nb == null) {
                    nbVal = nbVal.replaceFirst("Behavior", "");
                    int n = new Integer(nbVal).intValue();
                    int numBehaviors = provider.getNodeBehaviors().length;
                    if (n >= numBehaviors) {
                        Class substitution = TemplateParser.
                        this.provider.getNodeBehaviors()[0];
                        TemplateParser.this.warn( "Template has more node " +
                        "types than you have loaded.\n" + substitution.getName() +
                        " has been substituted for the missing node type(s).");
                        nb = substitution;
                    } else {
                        nb = provider.getNodeBehaviors()[n];
                    }
                }
                
                Node n = new Node(nb, x, y, name, show, enabled);
                nodeMap.put(id, n);
                TemplateParser.this.elements.add(n);
            }
            
            if ("channel".equals(qname)) {
                String startVal = attributes.getValue("start");
                String endVal = attributes.getValue("end");
                String latencyVal = attributes.getValue("latency");
                String capacityVal = attributes.getValue("capacity");
                String enabledVal = attributes.getValue("enabled");
                
                int latency = new Integer(latencyVal).intValue();
                int capacity = new Integer(capacityVal).intValue();
                boolean enabled = new Boolean(enabledVal).booleanValue();
                
                startVal = startVal.replaceFirst("node", "");
                Integer startID = new Integer(startVal);
                Node start = (Node) nodeMap.get(startID);
                
                endVal = endVal.replaceFirst("node", "");
                Integer endID = new Integer(endVal);
                Node end = (Node) nodeMap.get(endID);
                
                Channel chan = null;
                try {
                    chan = new Channel(start, end, enabled,
                    latency, capacity);
                } catch (SameNodeException sne) {
                    throw new SAXException(sne);
                }
                TemplateParser.this.elements.add(chan);
            }
            
        }
    }
}
