/*
 * Registrar.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu
 *
 * Created on February 20, 2003, 10:29 AM
 */

package nodenet.registrar;

import nodenet.Node;
import java.beans.VetoableChangeListener;

/**
 * A Registrar is an object that can track and validate the names of nodes.
 * A registrar can also provide a legal name for a node. Current Node Names
 * can be both registered and unregistered so that names can be reused if
 * nodes are deleted. The  * registrar may also provide access to a listing
 * of registered names, and access to a node via it's name, but
 * implementations may choose to throw an
 * <code>UnsupportedOperationException</code> instead of providing these
 * facilities. <p>
 *
 * Implicit in the design of this interface is the assumed existance of a
 * <code>public String getNodeName()</code> and a
 * <code>public void setNodeName(String)</code> or similar method on
 * <code>Node</code> objects.
 *
 * @author  Patrick G, Heck, gus.heck@olin.edu
 * @version $Id: NodeNameRegistrar.java,v 1.5 2004/01/14 21:43:17 gus Exp $
 */
public interface NodeNameRegistrar extends VetoableChangeListener {
    
    /**
     * Tests a name to see if it is currently legal. The legality of the name is
     * not guaranteed to remain invariant for all time. This is particularly
     * true if the policy implemented by the <code>Registrar</code> requires
     * names to be unique.
     *
     * @param   aName   The node name to be tested.
     * @return <code>true</code> if the name is legal, <code>false</code>
     *          otherwise.
     */
    public boolean isLegalName(String aName);
    
    /**
     * Provides a name that is currently legal. The legality of the name is
     * not guaranteed to remain legal for all time. This is particularly true
     * if the policy implemented by the <code>Registrar</code> requires names
     * to be unique. The implementation of this method should make a reasonable
     * attempt not to give out names that will become illegal quickly, either
     * due to temporal relevance of the name, or due to copetition between
     * names given out to the clients. For example, this method should not
     * give out the same node name on consecutive invocations if uniqueness is
     * a qualification for legality.
     *
     * @return A name that is currently legal, and should remain legal for a
     *          reasonable period of time, if possible.
     */
    public String getLegalName();
    
    /**
     * Provides a list of currently registered node names. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus list or null.
     *
     * @return An array of node names currently in use.
     */
    public String[] getRegisteredNames() throws UnsupportedOperationException;
    
    /**
     * Provides access to nodes via their registred name. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus node or null.
     *
     * @param aName   The name of the node to look for.
     * @return        A reference to the node if found, null otherwise.
     */
    public Node getNode(String aName) throws UnsupportedOperationException;
    
    /**
     * Attempt to register a node. This method proposes to register a node
     * with an existing name. The node will be rejected if and only if it
     * contains a name that would currently generate a false response
     * from {@link #isLegalName(String)}. If the name contained in the node
     * is rejected, The node is not registered or tracked in any way and an
     * <code>IllegalArgumentException</code> will be thrown.
     *
     * @param aNode The node that wishes to register itself.
     */
    public void register(Node aNode) throws IllegalArgumentException;
    
    /**
     * Ensures that the specified node is nolonger registered. If the node
     * was never registred, it simply returns with no side effects.
     *
     * @param aNode the node to unregister
     */
    public void unregister(Node aNode);
    
    /**
     * Tests to see if a node has been registered.
     *
     * @param aNode The node to test
     * @return <code>true</code> if the node is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(Node aNode);
    /**
     * Tests to see if a name belongs to a registered node. For obvious reasons
     * the results of this test are only valid while the client has a lock
     * on the registry.
     *
     * @param aName The node name to test
     * @return <code>true</code> if the name is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(String aName);
    
    /**
     * Gives the node a legal name and registers the node in one step. This
     * method guarantees that the node will be registered, and no exceptions
     * should be thrown. The node to be registered must expose a method
     * with the signature setName(String).
     */
    public void nameAndRegister(Node aNode);
    
}

/*
 * $Log: NodeNameRegistrar.java,v $
 * Revision 1.5  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.4  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.3  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.2  2003/02/25 22:35:18  gus
 * adding log comment
 *
 */
