/*
 * NodeNameRegistrar.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Created on February 19, 2003, 2:32 PM
 */

package nodenet.registrar;

import nodenet.Node;
import nodenet.NodeConfigurationBean;

import java.io.Serializable;
import java.io.ObjectStreamException;
import java.util.List;
import java.util.Vector;
import java.beans.VetoableChangeListener;
import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;

/**
 * Tracks all nodes and ensures that they have legal names. Besides providing
 * legal names via the {@link #nameAndRegister} method it listens for changes to
 * node names and vetos any attempt to create am illegal node name. To avoid
 * multiple registry problems, this class is uses the sigleton pattern, and
 * the one true <code>NodeNameRegistrar</code> can be obtained via the
 * static method {@link #getInstance()}.
 *
 * <p>This implementation only checks for uniquenes of names, and
 * accomodates up to Integer.MAX_VALUE nodes, wich is
 * limited by the Vector class used to store node references. There is also
 * a limit of Integer.MAX_VALUE name requests during the lifetime of this
 * registrar object, as unique names are generated in the format "Node X"
 * where X is replaced by an integer. <strong>If a node name is changed to
 * "Node Y" where Y is an integer.
 *
 * @author  Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: NodeNameRegistry.java,v 1.5 2004/01/14 21:04:16 gus Exp $
 */
public final class NodeNameRegistry implements NodeNameRegistrar {
    
    private static final NodeNameRegistry SINGLETON = new NodeNameRegistry();
    
    private NodeNameRegistrar policy;
    
    /** Prevent independant instantiation */
    private NodeNameRegistry() {
    }
    
    /**
     * Factory method to get the one true <code>NodeNameRegistrar</code>.
     *
     * @return the registrar object.
     */
    public static NodeNameRegistry getInstance() {
        return SINGLETON;
    }
    
    /**
     * Set the registrartion policy one time only. This method may not be
     * invoked more than once. Subsequent invocations will throw an
     * <code>IllegalStateException</code>. This is necessary to prevent
     * the registered nodes from being lost. In the future a changePolicy
     * method may be provided for "graceful" transition between policies.
     * Note that "graceful" may include deletion of existing nodes.
     *
     * @param policy  An naming policy implemented as a NodeNameRegistrar.
     */
    public void setPolicy(NodeNameRegistrar policy) {
        if (!isPolicySet()) {
            this.policy = policy;
        } else {
            throw new IllegalStateException("Policy already set!");
        }
    }
    
    /**
     * Check to see if the policy has been set for this registry.
     *
     * @return <code>true</code> if the policy has been set, <code>false</code>
     *         otherwise
     */
    
    public boolean isPolicySet() {
        return (policy == null) ? false : true;
    }
    
    // This allows us to deserialize objects containing a reference to the
    // NodeNameRegistry and never cause duplicate registries. This technique
    // means that when nodes are deserialized they need to re-register
    // with the existing registry.
    private Object readResolve() throws ObjectStreamException {
        return SINGLETON;
    }
    
    public void vetoableChange(PropertyChangeEvent pce) throws PropertyVetoException {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set! (method:vetoableChange)");
        }
        
        if (pce.getPropertyName().equals(NodeConfigurationBean.PROP_NODE_NAME)) {
            this.policy.vetoableChange(pce);
        }
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides a name that is currently legal. The legality of the name is
     * not guaranteed to remain legal for all time. This is particularly true
     * if the policy implemented by the <code>Registrar</code> requires names
     * to be unique. The implementation of this method should make a reasonable
     * attempt not to give out names that will become illegal quickly, either
     * due to temporal relevance of the name, or due to copetition between
     * names given out to the clients. For example, this method should not
     * give out the same node name on consecutive invocations if uniqueness is
     * a qualification for legality.
     *
     * @returns A name that is currently legal, and should remain legal for a
     *          reasonable period of time, if possible.
     */
    public String getLegalName() {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:getLegalName)");
        }
        return this.policy.getLegalName();
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides access to nodes via their registred name. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus node or null.
     *
     * @param aName   The name of the node to look for.
     * @return        A reference to the node if found, null otherwise.
     */
    public Node getNode(String aName) throws UnsupportedOperationException {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:getNode(String))");
        }
        return this.policy.getNode(aName);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides a list of currently registered node names. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus list or null.
     *
     * @returns An array of node names currently in use.
     */
    public String[] getRegisteredNames() throws UnsupportedOperationException {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:getRegisteredNames)");
        }
        return this.policy.getRegisteredNames();
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests a name to see if it is currently legal. The legality of the name is
     * not guaranteed to remain invariant for all time. This is particularly
     * true if the policy implemented by the <code>Registrar</code> requires
     * names to be unique.
     *
     * @param   aName   The node name to be tested.
     * @returns <code>true</code> if the name is legal, <code>false</code>
     *          otherwise.
     */
    public boolean isLegalName(String aName) {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:isLegalName)");
        }
        return this.policy.isLegalName(aName);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests to see if a name belongs to a registered node. For obvious reasons
     * the results of this test are only valid while the client has a lock
     * on the registry.
     *
     * @param The node name to test
     * @return <code>true</code> if the name is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(String aName) {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:isRegistered)");
        }
        return this.policy.isRegistered(aName);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests to see if a node has been registered.
     *
     * @param The node to test
     * @return <code>true</code> if the node is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(Node aNode) {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:isRegistered)");
        }
        return this.policy.isRegistered(aNode);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Gives the node a legal name and registers the node in one step. This
     * method guarantees that the node will be registered, and no exceptions
     * should be thrown. The node to be registered must expose a method
     * with the signature setName(String).
     */
    public void nameAndRegister(Node aNode) {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:nameAndRegister)");
        }
        this.policy.nameAndRegister(aNode);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Attempt to register a node. This method proposes to register a node
     * with an existing name. The node will be rejected if and only if it
     * contains a name that would currently generate a false response
     * from {@link #isLegalName(String)}. If the name contained in the node
     * is rejected, The node is not registered or tracked in any way and an
     * <code>IllegalArgumentException</code> will be thrown.
     *
     * @param aNode The node that wishes to register itself.
     */
    public void register(Node aNode) throws IllegalArgumentException {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:register)");
        }
        this.policy.register(aNode);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Ensures that the specified node is nolonger registered. If the node
     * was never registred, it simply returns with no side effects.
     *
     * @param aNode the node to unregister
     */
    public void unregister(Node aNode) {
        if (!isPolicySet()) {
            throw new IllegalStateException("Policy not set!(method:unregister)");
        }
        this.policy.unregister(aNode);
    }
    
}

/*
 * $Log: NodeNameRegistry.java,v $
 * Revision 1.5  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.4  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.3  2003/02/25 21:42:31  gus
 * Improve error messages to specify which method generated them
 *
 * Revision 1.2  2003/02/24 15:54:01  gus
 * Add support for configuring the destination of generated packets
 *
 * Revision 1.1  2003/02/21 17:47:06  gus
 * A set of classes to support name checking  and generation schemes for nodes.
 *
 */