/*
 * UniqueNodeNamePolicy.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Created on February 20, 2003, 11:54 AM
 */

package nodenet.registrar;

import nodenet.Node;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyVetoException;

import java.util.Vector;
import java.util.ArrayList;

/**
 * This class is designed to be plugged into {@link NodeNameRegistry}
 * via the <code>setPolicy(NodeNameRegistrar)</code> method. It enforces
 * the uniqueness of node names, and provides access to a node via its
 * textual name.
 *
 * @author  Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: UniqueNodeNamePolicy.java,v 1.8 2004/01/14 21:43:17 gus Exp $
 */
public class UniqueNodeNamePolicy implements NodeNameRegistrar {
    
    private static final String NAME_PREFIX = "Node ";
    private volatile int nodeSerialNum;
    
    private Vector nodes = new Vector();
    
    /** Creates a new instance of UniqueNodeNamePolicy */
    public UniqueNodeNamePolicy() {
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides a name that is currently legal. The legality of the name is
     * not guaranteed to remain legal for all time. This is particularly true
     * if the policy implemented by the <code>Registrar</code> requires names
     * to be unique. The implementation of this method should make a reasonable
     * attempt not to give out names that will become illegal quickly, either
     * due to temporal relevance of the name, or due to copetition between
     * names given out to the clients. For example, this method should not
     * give out the same node name on consecutive invocations if uniqueness is
     * a qualification for legality.
     *
     * @returns A name that is currently legal, and should remain legal for a
     *          reasonable period of time, if possible.
     */
    public synchronized String getLegalName() {
        String temp;
        do {
            nodeSerialNum++;
            temp = NAME_PREFIX + nodeSerialNum;
        } while (!isUnique(temp));
        
        return temp;
    }
    
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides access to nodes via their registred name. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus node or null.
     *
     * @param aName   The name of the node to look for.
     * @return        A reference to the node if found, null otherwise.
     */
    public synchronized Node getNode(String aName) throws UnsupportedOperationException {
        for (int i=0; i<nodes.size(); i++) {
            Node aNode = ((Node)nodes.get(i));
            if(aNode.getName().equals(aName)) {
                return aNode;
            }
        }
        return null;
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Provides a list of currently registered node names. This method is
     * optional, but if it is not implemented it should throw an
     * <code>UnsupportedOperationException</code> rather than returning
     * a bogus list or null.
     *
     * @returns An array of node names currently in use.
     */
    public synchronized String[] getRegisteredNames() throws UnsupportedOperationException {
        String[] temp = new String[nodes.size()];
        for (int i=0; i<temp.length;i++) {
            temp[i] = ((Node)nodes.get(i)).getName();
        }
        return temp;
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests a name to see if it is currently legal. The legality of the name is
     * not guaranteed to remain invariant for all time. This is particularly
     * true if the policy implemented by the <code>Registrar</code> requires
     * names to be unique.
     *
     * @param   aName   The node name to be tested.
     * @returns <code>true</code> if the name is legal, <code>false</code>
     *          otherwise.
     */
    public boolean isLegalName(String aName) {
        return isUnique(aName);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests to see if a name belongs to a registered node. For obvious reasons
     * the results of this test are only valid while the client has a lock
     * on the registry.
     *
     * @param The node name to test
     * @return <code>true</code> if the name is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(String aName) {
        String[] names = getRegisteredNames();
        for (int i=0; i<names.length; i++) {
            if (names[i].equals(aName)) {
                return true;
            }
        }
        return false;
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Tests to see if a node has been registered.
     *
     * @param The node to test
     * @return <code>true</code> if the node is registered <code>false</code>
     *         otherwise.
     */
    public boolean isRegistered(Node aNode) {
        return isRegistered(aNode.getName());
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Gives the node a legal name and registers the node in one step. This
     * method guarantees that the node will be registered, and no exceptions
     * should be thrown. The node to be registered must expose a method
     * with the signature setName(String).
     */
    public synchronized void nameAndRegister(Node aNode) {
        aNode.setName(getLegalName());
        this.register(aNode);
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Attempt to register a node. This method proposes to register a node
     * with an existing name. The node will be rejected if and only if it
     * contains a name that would currently generate a false response
     * from {@link #isLegalName(String)}. If the name contained in the node
     * is rejected, The node is not registered or tracked in any way and an
     * <code>IllegalArgumentException</code> will be thrown.
     *
     * @param aNode The node that wishes to register itself.
     */
    public synchronized void register(Node aNode) throws IllegalArgumentException {
        
        if (!isUnique(aNode.getName())) {
            throw new IllegalArgumentException("Tried to register " +
            "a node with a name that exists!\n Use nameNode to " +
            "get a unique name before registering.\n");
        } else {
            nodes.add(aNode);
        }
    }
    
    /* (Copy of original doc. Generated version is inherited)
     *
     * Ensures that the specified node is nolonger registered. If the node
     * was never registred, it simply returns with no side effects.
     *
     * @param aNode the node to unregister
     */
    public void unregister(Node aNode) {
        nodes.remove(aNode);
    }
    
    public void vetoableChange(PropertyChangeEvent pce) throws PropertyVetoException {
        if (!isLegalName((String)pce.getNewValue())) {
            throw new PropertyVetoException("Duplicate names not allowed!",pce);
        }
    }
    
    
    /**
     * Checks for the uniqueness of the a potential node name.
     *
     * @param name The name to check for uniqueness.
     * @return     <code>true</code> if the name is not in use <code>false</code>
     *             otherwise
     */
    private synchronized boolean isUnique(String name) {
        for (int i=0; i<nodes.size();i++){
            if (name.equals(((Node) nodes.get(i)).getName())) {
                return false;
            }
        }
        return true;
    }
}


/*
 * $Log: UniqueNodeNamePolicy.java,v $
 * Revision 1.8  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.7  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.6  2003/02/25 22:35:18  gus
 * adding log comment
 *
 */