/*
 * NodeNetFrame.java
 *
 * Developed for the "Rethinking CS101" project. See http://www.cs101.org, the
 * CS101 homepage or email las@olin.edu.
 *
 * Please do not redistribute without obtaining permission.
 */

package nodenet;

import org.xml.sax.SAXException;

import java.io.File;
import java.io.FileWriter;
import java.io.IOException;
import java.io.FileNotFoundException;

import java.awt.AWTEvent;
import java.awt.Dimension;

import java.awt.event.ActionEvent;
import java.awt.event.WindowEvent;
import java.awt.event.WindowAdapter;
import java.awt.event.ComponentEvent;
import java.awt.event.ActionListener;
import java.awt.event.ComponentAdapter;

import java.net.URL;
import java.net.URLClassLoader;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;

import java.util.Iterator;
import java.util.HashSet;

import javax.xml.parsers.ParserConfigurationException;

import javax.swing.JMenu;
import javax.swing.JFrame;
import javax.swing.JMenuBar;
import javax.swing.JMenuItem;
import javax.swing.JSplitPane;
import javax.swing.JOptionPane;
import javax.swing.JFileChooser;


/**
 * The top level frame for nodeNet.  Creates a SimulationPanel and a
 * NodeTypeSelector, sets up menus, and leaves the rest to the event
 * handlers.<p>
 *
 * @see nodenet.SimulationPanel
 * @see nodenet.NodeTypeSelector
 *
 * @author Todd C. Parnell, tparnell@ai.mit.edu
 * @author Patrick G. Heck, gus.heck@olin.edu
 * @version $Id: NodeNetFrame.java,v 1.23 2004/02/09 19:17:49 gus Exp $
 */
public class NodeNetFrame extends JFrame {
    
    private SimulationPanel sp;
    private JSplitPane jsp;
    private NodeTypeSelector nts;
    
    // this class loader must be found by the Main thread so we have to
    // cache it for later use. Otherwise it won't work in Java WebStart.
    private ClassLoader cl = Thread.currentThread().getContextClassLoader();
    
    private URLClassLoader uLoad = URLClassLoader.newInstance(new URL[] {}, cl);
    private static final int MIN_X = 250;
    private static final int MIN_Y = 150;
    
    /**
     * Constructs the GUI.
     *
     * @param nodeBehaviors a (possibly empty) list of NodeBehaviors to
     * add to the default list of NodeBehaviors.  Typically a student's
     * class.
     */
    public NodeNetFrame(String[] nodeBehaviors) {
        super("Network Simulation");
        
        this.nts = new NodeTypeSelector();
        this.sp = new SimulationPanel();
        this.nts.addPropertyChangeListener(this.sp);
        this.jsp = new JSplitPane(JSplitPane.HORIZONTAL_SPLIT,
        this.nts, this.sp);
        this.jsp.setResizeWeight(0);
        this.getContentPane().add(this.jsp);
        
        // try to keep the window from getting too small, and the 
        // JSeparator from getting stuck on the right side.
        this.addComponentListener( new ComponentAdapter() {
            public void componentResized(ComponentEvent ce) {
                //System.out.println(ce.paramString());
                int width = NodeNetFrame.this.getWidth();
                int height = NodeNetFrame.this.getHeight();
                if (width < MIN_X || height < MIN_Y) {
                    // There are serious timing issues with resize events.
                    // This delay attempts to make us the last event sent. 
                    // There is also some sort of timeout for event processing 
                    // which is about 2-3 seconds long the first time, and then 
                    // seems to drop to less than a second. This is probably 
                    // awt trying to keep alive when it has too many events or 
                    // an event is blocking. The result is that if you resize 
                    // to a very small size, and do not release the mouse 
                    // button for for 4-5 sec the size enforcment breaks 
                    // because the event queue starts dropping resize events 
                    // quickly without waiting for the user to release the 
                    // mouse button and give control back to Java. This results
                    // in java thinking it has updated the window size when in
                    // fact it has not. increasing the delay below makes the
                    // resize more sluggish, but more reliable. For now we just
                    // have to live with this hack I think.
                    try {Thread.sleep(500);} catch (InterruptedException ie) {}
                    NodeNetFrame.this.setSize(1,1);
                    try {Thread.sleep(50);} catch (InterruptedException ie) {}
                    NodeNetFrame.this.setSize(
                    width < MIN_X ? MIN_X : width,
                    height < MIN_Y ? MIN_Y : height);
                    
                    jsp.setDividerLocation(0.5);
                    //System.out.println("resized to " + NodeNetFrame.this.getSize());
                    
                }
            }
        });
        
        JMenuBar menubar = new JMenuBar();
        this.setJMenuBar(menubar);
        JMenu menu = new JMenu("File");
        menu.setMnemonic('f');
        
        menubar.add(menu);
        
        JMenuItem menuitem = new JMenuItem("Load Behavior...");
        menuitem.setMnemonic('l');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                String temp = "";
                try {
                    JOptionPane jop = new JOptionPane();
                    
                    temp = JOptionPane.showInputDialog("Enter a class that " +
                    "is on the classpath");
                    if (temp == null || "".equals(temp)) {
                        return;
                    }
                    Class clazz;
                    synchronized (uLoad) {
                        clazz = NodeNetFrame.this.nts.loadBehavior(temp,
                        NodeNetFrame.this.uLoad);
                    }
                    if (clazz == null) {
                        throw new ClassNotFoundException();
                    }
                    NodeNetFrame.this.nts.setNodeBehaviors(
                    NodeTypeSelector.NEXT, clazz);
                } catch (java.security.AccessControlException ace) {
                    // we're running as an applet
                    NodeNetFrame.this.showAppletWarning();
                } catch (ClassNotFoundException cnfe) {
                    String temp2 = "nodenet." + temp;
                    try {
                        Class clazz;
                        synchronized (uLoad) {
                            clazz = NodeNetFrame.this.nts.loadBehavior(temp2,
                            NodeNetFrame.this.uLoad);
                        }
                        if (clazz == null) {
                            throw new ClassNotFoundException();
                        }
                        NodeNetFrame.this.nts.setNodeBehaviors(
                        NodeTypeSelector.NEXT, clazz);
                    } catch (ClassNotFoundException cnfe2) {
                        StringBuffer err = new StringBuffer();
                        err.append("Unable to find the class '");
                        err.append(temp + "' or 'nodenet." + temp +"'." );
                        err.append("\nExpanded classpath URLs:");
                        URL[] foobar = NodeNetFrame.this.uLoad.getURLs();
                        for (int i = 0; i < foobar.length; i++) {
                            err.append("\n" + foobar[i].toString());
                        }
                        System.err.println(err);
                        JOptionPane.showMessageDialog(null, err);
                    } catch (IllegalArgumentException iae) {
                        
                        System.err.println("Error instantiating class '" +
                        temp + "' or nodenet." + temp + "'.");
                        URL[] foobar = NodeNetFrame.this.uLoad.getURLs();
                        for (int i = 0; i < foobar.length; i++) {
                            System.err.println(foobar[i]);
                        }
                        iae.printStackTrace();
                    }
                } catch (Exception ex) {
                    JOptionPane.showMessageDialog(null,ex);
                }
            }
        });
        menu.add(menuitem);
        menuitem = new JMenuItem("Expand Classpath...");
        menuitem.setMnemonic('x');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                String temp = "";
                try {
                    JFileChooser chooser = new JFileChooser();
                    chooser.setFileFilter(
                    new javax.swing.filechooser.FileFilter() {
                        public String getDescription() {
                            return "Java Archive (*.jar) or Directory";
                        }
                        public boolean accept(File f) {
                            return
                            f.getName().matches(".*\\.jar") || f.isDirectory();
                        }
                    });
                    chooser.setFileSelectionMode(
                    JFileChooser.FILES_AND_DIRECTORIES);
                    chooser.setMultiSelectionEnabled(true);
                    int returnVal = chooser.showOpenDialog( NodeNetFrame.this );
                    File[] files = chooser.getSelectedFiles();
                    URL[] prev = uLoad.getURLs();
                    URL[] newURLs = new URL[prev.length + files.length];
                    for (int i=0; i < files.length; i++) {
                        //System.out.print("new:" + files[i]);
                        newURLs[i] = files[i].toURL();
                    }
                    for (int i=files.length; i < newURLs.length; i++) {
                        //System.out.println("old:" + prev[i - files.length]);
                        newURLs[i] = prev[i - files.length];
                    }
                    synchronized (uLoad) {
                        uLoad = new URLClassLoader(newURLs, cl);
                    }
                    
                } catch (java.security.AccessControlException ace) {
                    // we're running as an applet
                    NodeNetFrame.this.showAppletWarning();
                } catch (Exception e2) {
                    e2.printStackTrace();
                }
            }
        });
        menu.add(menuitem);
        
        menu.addSeparator();
        
        menuitem = new JMenuItem("Open a Template...");
        menuitem.setMnemonic('o');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                TemplateParser tparse = null;
                try {
                    JFileChooser chooser = new JFileChooser();
                    chooser.setFileFilter(
                    new javax.swing.filechooser.FileFilter() {
                        public String getDescription() {
                            return "Node Net Template File (*.nnt)";
                        }
                        public boolean accept(File f) {
                            return
                            f.getName().matches(".*\\.nnt") || f.isDirectory();
                        }
                    });
                    int returnVal = chooser.showOpenDialog( NodeNetFrame.this );
                    if (returnVal != JFileChooser.APPROVE_OPTION ||
                    chooser.getSelectedFile() == null) return;
                    String file = chooser.getCurrentDirectory() +
                    System.getProperty("file.separator") +
                    chooser.getSelectedFile().getName();
                    
                    tparse =
                    new TemplateParser(
                    NodeNetFrame.this.sp.getBehaviorSource());
                    
                    NodeNetFrame.this.sp.clear();
                    NodeNetFrame.this.sp.setBsElements(tparse.parse(file));
                    
                } catch (java.security.AccessControlException ace) {
                    // we're running as an applet
                    NodeNetFrame.this.showAppletWarning();
                } catch (SAXException saxe) {
                    // print out the error
                    saxe.printStackTrace();
                    JOptionPane.showMessageDialog(null, saxe);
                } catch (ParserConfigurationException pce) {
                    // print out the error
                    pce.printStackTrace();
                    JOptionPane.showMessageDialog(null, pce);
                } catch (FileNotFoundException fnfe) {
                    // print out the error
                    JOptionPane.showMessageDialog(null,"File Not Found!");
                } catch (IOException ioe) {
                    // print out the error
                    ioe.printStackTrace();
                    JOptionPane.showMessageDialog(null, ioe);
                } catch (Throwable t) { 
                    // display anything else that goes wrong.
                    JOptionPane.showMessageDialog(null, t);
                } finally {
                    if (tparse != null) {
                        // Identical warnings should be coalesced into a single
                        // warning to avoid dialog spam.
                        HashSet s = new HashSet(tparse.getWarnings());
                        for (Iterator iter = s.iterator(); iter.hasNext();) {
                            String message = (String) iter.next();
                            JOptionPane.showMessageDialog(null, message);
                        }
                    }
                }
            }
        });
        
        menu.add(menuitem);
        
        menuitem = new JMenuItem("Save Template As...");
        menuitem.setMnemonic('s');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                try {
                    JFileChooser chooser = new JFileChooser();
                    
                    chooser.setFileFilter(
                    new javax.swing.filechooser.FileFilter() {
                        public String getDescription() {
                            return "Node Net Template file (.nnt)";
                        }
                        public boolean accept(File f) {
                            return
                            f.getName().matches(".*\\.nnt") || f.isDirectory();
                        }
                    });
                    int returnVal = chooser.showSaveDialog(NodeNetFrame.this);
                    if (returnVal != JFileChooser.APPROVE_OPTION ||
                    chooser.getSelectedFile() == null) return;
                    String file = chooser.getCurrentDirectory() +
                    System.getProperty("file.separator") +
                    chooser.getSelectedFile().getName();
                    if (file.indexOf(".") == -1) {
                        file = file + ".nnt";
                    }
                    
                    FileWriter fw = new FileWriter(file);
                    String xml = NodeNetFrame.this.sp.templateXML();
                    fw.write(xml);
                    fw.flush();
                    fw.close();
                } catch (java.security.AccessControlException ace) {
                    // we're running as an applet
                    NodeNetFrame.this.showAppletWarning();
                } catch (Exception e2) {
                    System.out.println(e2);
                }
            }
        });
        menu.add(menuitem);
        
        menu.addSeparator();
        
        menuitem = new JMenuItem("Quit");
        menuitem.setMnemonic('q');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                try {
                    System.exit(0);
                } catch (java.security.AccessControlException ace) {
                    NodeNetFrame.this.showAppletWarning();
                }
            }
        });
        menu.add(menuitem);
        
        menu = new JMenu("Simulation");
        menu.setMnemonic('s');
        menubar.add(menu);
        
        menuitem = new JMenuItem("Start");
        menuitem.setMnemonic('s');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                NodeNetFrame.this.sp.startSimulation();
            }
        });
        menu.add(menuitem);
        
        menuitem = new JMenuItem("Pause");
        menuitem.setMnemonic('p');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                NodeNetFrame.this.sp.stopSimulation();
            }
        });
        menu.add(menuitem);
        
        menuitem = new JMenuItem("Reset");
        menuitem.setMnemonic('r');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                NodeNetFrame.this.sp.reset();
            }
        });
        menu.add(menuitem);
        
        menu.addSeparator();
        
        menuitem = new JMenuItem("Clear");
        menuitem.setMnemonic('c');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                NodeNetFrame.this.sp.clear();
            }
        });
        menu.add(menuitem);
        
        
        menu = new JMenu("Selection");
        menu.setMnemonic('e');
        menubar.add(menu);
        
        menuitem = new JMenuItem("Enable");
        menuitem.setMnemonic('e');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                sp.setSelectionEnabled(true);
            }
        });
        menu.add(menuitem);
        
        menuitem = new JMenuItem("Disable");
        menuitem.setMnemonic('d');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                sp.setSelectionEnabled(false);
            }
        });
        menu.add(menuitem);
        
        menuitem = new JMenuItem("Configure");
        menuitem.setMnemonic('c');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                sp.configureSelection();
            }
        });
        menu.add(menuitem);
        
        menu.addSeparator();
        
        menuitem = new JMenuItem("Remove");
        menuitem.setMnemonic('r');
        menuitem.addActionListener(new ActionListener() {
            public void actionPerformed(ActionEvent e) {
                sp.deleteSelection();
            }
        });
        menu.add(menuitem);
        
        this.addWindowListener(new WindowAdapter() {
            public void windowClosing(WindowEvent e) {
                try {
                    System.exit(0);
                } catch (java.security.AccessControlException ace) {
                }
            }
        });
        
        this.pack();
        this.jsp.setDividerLocation(this.nts.getWidth());
        this.show();
    }
    
    /**
     * Pops up a warning dialog.
     */
    private void showAppletWarning() {
        Object[] options = { "Cancel" };
        JOptionPane.showOptionDialog(
        null,
        "This action not available when running as an applet.",
        "Network Simulation",
        JOptionPane.DEFAULT_OPTION,
        JOptionPane.WARNING_MESSAGE,
        null, options, options[0]);
    }
}

/*
 * $Log: NodeNetFrame.java,v $
 * Revision 1.23  2004/02/09 19:17:49  gus
 * Remove debugging JOptionPane
 *
 * Revision 1.22  2004/02/09 17:22:25  gus
 * more debugging stuff to track down web start problem.
 *
 * Revision 1.21  2004/02/05 00:00:55  gus
 * more error info in graphical only environments
 *
 * Revision 1.20  2004/02/04 22:17:44  gus
 * helpful comment (I hope)
 *
 * Revision 1.19  2004/02/04 21:57:30  gus
 * Better error reporting
 *
 * Revision 1.18  2004/02/04 21:41:13  gus
 * slightly more robust classloader... might work under webstart
 *
 * Revision 1.17  2004/01/26 18:19:38  gus
 * remove prints and just live with the imperfect hack.
 *
 * Revision 1.16  2004/01/16 20:48:59  gus
 * Allow user to add jars and dirs to the classpath by using a custom
 * ClassLoader.
 * Also (try to) prevent the window from getting too small since funny things
 * happen with the JSeparator and very small window sizes.
 *
 * Revision 1.15  2004/01/14 23:42:35  gus
 * specifying the encoding is clearly causing more problems that it is worth
 *
 * Revision 1.14  2004/01/14 23:28:11  gus
 * another attempt to get the encoding right
 *
 * Revision 1.13  2004/01/14 23:19:17  gus
 * Make sure the encoding matches the file, rather than assuming utf-8
 *
 * Revision 1.12  2004/01/14 21:43:17  gus
 * more javadoc, plus reformat
 *
 * Revision 1.11  2004/01/14 21:04:16  gus
 * More javadoc fixes
 *
 * Revision 1.10  2004/01/14 20:23:21  gus
 * Javadoc and comment cleanup
 *
 * Revision 1.9  2004/01/13 19:35:27  gus
 * Simulation Panel, channel and node are all Resettable now.
 * CountingNodeBehavior refactored into Counter
 *
 * Revision 1.8  2004/01/09 20:02:33  gus
 * Clean up the loading of classes from the classpath. We should not be using a file
 * dialog. We are not selecting files. Replaced with JOptionPane.
 *
 * Revision 1.7  2004/01/09 18:15:01  gus
 * Account for more types in template than loaded behaviors.
 * Tell user when file not found.
 * Filter template dialogs to *.nnt
 *
 * Revision 1.6  2004/01/08 22:48:08  gus
 * Can now read xml template file (only tested breifly)
 *
 * Revision 1.5  2004/01/07 20:10:11  gus
 * Update comments to refelect disappearance of ControlPanel Class
 *
 * Revision 1.4  2004/01/07 19:51:59  gus
 * Put the buttons in a scroll pane so that the JSplitPane is useful again
 *
 * Revision 1.3  2004/01/07 18:32:47  gus
 * Ripped out ControlPanel and NodeNetRadioButton.
 * Replaced them with NodeTypeSelector, which provides
 * The loaded User NodeBehavior types as an indexed
 * property. Simulation panel now implements PropertyChangeListener
 * to listen for changes in the selected node behavior. The
 * buttons in the display are now JToggleButtons, and will
 * use the public field BEHAVIOR_NAME to lable the buttons
 * if it is declared for any NodeBehavior.
 *
 * Revision 1.2  2002/08/20 15:41:09  gus
 * corrected package name
 *
 * Revision 1.1  2002/06/13 17:33:21  gus
 * Moved all java files into the nodenet directory (who let them out anyway?)
 *
 * Revision 1.1.1.1  2002/06/05 21:56:35  root
 * CS101 comes to Olin finally.
 *
 * Revision 1.4  2000/05/09 06:03:54  mharder
 * Changed packagename from nodeNet to nodenet.
 *
 * Revision 1.3  1999/08/04 09:08:54  jsmthng
 * Added javadoc comments to InputChannelVector and OutputChannelVector;
 * finished updating the rest of the nodeNet package to reflect new
 * changes in name and code.
 *
 * Modified index.html to reflect the new nodeNet code, as well as to
 * clarify some parts of the problem set.
 *
 * Revision 1.1  1999/07/30 01:09:23  jsmthng
 * Renaming BinSort package to nodeNet; moving directories and files as
 * necessary.
 *
 * Revision 1.5  1999/01/25 19:31:49  tparnell
 * Fixed packages.  cs101.util.Queue is now an interface, replaced with DefaultQueue
 *
 * Revision 1.4  1999/01/21 20:57:03  tparnell
 * it worked
 *
 * Revision 1.3  1998/08/14 16:17:43  tparnell
 * now appearing in applet version
 *
 * Revision 1.2  1998/08/12 19:29:26  tparnell
 * Another pass after comments from las & natashao.  Added support to
 * dynamically add NodeBehaviors.  Add keyboard shortcuts to menus.  Added
 * workaround to jdk bug relating to lighweight components.  Misc other
 * bugfixes.
 *
 * Revision 1.3  1998/08/10 17:45:50  tparnell
 * Revision to use JDK1.2 and Swing.  Redesign of GUI.  Removed old kludge
 * for file I/O and replaced with object serialization.  Channel no longer
 * requires animacy.  Removed unnessary dependencies between classes.
 * Added ability to configure channel's latency and capacity.  Added
 * javadoc to all files.  General code cleanup.
 *
 */
